r"""internal module used to get the list of default proxies.

This is an internal module and not intended to be used by Python code outside of
this package.

"""

__defaultProxiesJSON = """
{
    "sources": [
        { "name": "AnnotateTime", "label": "Annotate Time" },
        { "name": "Cone" },
        { "name": "Sphere" },
        { "name": "Text" },
        { "name": "Wavelet" }
    ],

    "filters": [
        { "name": "Calculator" },
        { "name": "CellDatatoPointData", "label": "Cell Data To Point Data" },
        { "name": "Clip" },
        { "name": "Contour" },
        { "name": "D3" },
        { "name": "ExtractCTHParts", "label": "Extract CTH Parts" },
        { "name": "ProcessIdScalars", "label": "Process ID Scalars" },
        { "name": "Reflect" },
        { "name": "Slice" },
        { "name": "StreamTracer", "label": "Stream Tracer" },
        { "name": "Threshold" },
        { "name": "Transform" },
        { "name": "Tube" },
        { "name": "Ribbon" },
        { "name": "WarpByScalar", "label": "Warp By Scalar" },
        { "name": "WarpByVector", "label": "Warp By Vector" },
        { "name": "ExtractBlock", "label": "Extract Blocks" }
    ],

    "readers": [
        { "name": "LegacyVTKReader", "extensions": [ "vtk" ], "method": "FileNames" },
        { "name": "Xdmf3ReaderS", "extensions": [ "xmf", "xdmf" ] }
    ]
}
"""

def getDefaultProxies():
    """Returns the JSON object for the default proxies configuration"""
    import json
    return json.loads(__defaultProxiesJSON)



from __future__ import absolute_import, division, print_function

import json, time, re

from vtk.vtkFiltersGeometry import vtkCompositeDataGeometryFilter
from vtk.vtkCommonCore import vtkTypeUInt32Array
from vtk.vtkWebCore import vtkWebApplication

from vtk.web import hashDataArray, getJSArrayType
from vtk.web import buffer, base64Encode

# -----------------------------------------------------------------------------
# Convenience class for caching data arrays, storing computed sha sums, keeping
# track of valid actors, etc...
# -----------------------------------------------------------------------------

class SynchronizationContext():
  def __init__(self, debug=False):
    self.dataArrayCache = {}
    self.lastDependenciesMapping = {}
    self.ingoreLastDependencies = False
    self.debugSerializers = debug
    self.debugAll = debug

  def setIgnoreLastDependencies(self, force):
    self.ingoreLastDependencies = force

  def cacheDataArray(self, pMd5, data):
    self.dataArrayCache[pMd5] = data

  def getCachedDataArray(self, pMd5, binary = False):
    cacheObj = self.dataArrayCache[pMd5]
    array = cacheObj['array']
    cacheTime = cacheObj['mTime']

    if cacheTime != array.GetMTime():
      if context.debugAll: print(' ***** ERROR: you asked for an old cache key! ***** ')

    if array.GetDataType() == 12:
      # IdType need to be converted to Uint32
      arraySize = array.GetNumberOfTuples() * array.GetNumberOfComponents()
      newArray = vtkTypeUInt32Array()
      newArray.SetNumberOfTuples(arraySize)
      for i in range(arraySize):
        newArray.SetValue(i, -1 if array.GetValue(i) < 0 else array.GetValue(i))
      pBuffer = buffer(newArray)
    else:
      pBuffer = buffer(array)

    if binary:
      # Convert the vtkUnsignedCharArray into a bytes object, required by Autobahn websockets
      return pBuffer.tobytes()

    return base64Encode(pBuffer)

  def checkForArraysToRelease(self, timeWindow = 20):
    cutOffTime = time.time() - timeWindow
    shasToDelete = []
    for sha in self.dataArrayCache:
      record = self.dataArrayCache[sha]
      array = record['array']
      count = array.GetReferenceCount()

      if count == 1 and record['ts'] < cutOffTime:
        shasToDelete.append(sha)

    for sha in shasToDelete:
      del self.dataArrayCache[sha]

  def getLastDependencyList(self, idstr):
    lastDeps = []
    if idstr in self.lastDependenciesMapping and not self.ingoreLastDependencies:
      lastDeps = self.lastDependenciesMapping[idstr]
    return lastDeps

  def setNewDependencyList(self, idstr, depList):
    self.lastDependenciesMapping[idstr] = depList

  def buildDependencyCallList(self, idstr, newList, addMethod, removeMethod):
    oldList = self.getLastDependencyList(idstr)

    calls = []
    calls += [ [addMethod, [ wrapId(x) ]] for x in newList if x not in oldList ]
    calls += [ [removeMethod, [ wrapId(x) ]] for x in oldList if x not in newList ]

    self.setNewDependencyList(idstr, newList)
    return calls

# -----------------------------------------------------------------------------
# Global variables
# -----------------------------------------------------------------------------

SERIALIZERS = {}
context = None

# -----------------------------------------------------------------------------
# Global API
# -----------------------------------------------------------------------------

def registerInstanceSerializer(name, method):
  global SERIALIZERS
  SERIALIZERS[name] = method

# -----------------------------------------------------------------------------

def serializeInstance(parent, instance, instanceId, context, depth):
  instanceType = instance.GetClassName()
  serializer = SERIALIZERS[instanceType] if instanceType in SERIALIZERS else None

  if serializer:
    return serializer(parent, instance, instanceId, context, depth)

  if context.debugSerializers:
    print('%s!!!No serializer for %s with id %s' % (pad(depth), instanceType, instanceId))

  return None

# -----------------------------------------------------------------------------

def initializeSerializers():
  # Actors/viewProps
  registerInstanceSerializer('vtkOpenGLActor', genericActorSerializer)
  registerInstanceSerializer('vtkPVLODActor', genericActorSerializer)

  # Mappers
  registerInstanceSerializer('vtkOpenGLPolyDataMapper', genericMapperSerializer)
  registerInstanceSerializer('vtkCompositePolyDataMapper2', genericMapperSerializer)

  # LookupTables/TransferFunctions
  registerInstanceSerializer('vtkLookupTable', lookupTableSerializer)
  registerInstanceSerializer('vtkPVDiscretizableColorTransferFunction', colorTransferFunctionSerializer)

  # Property
  registerInstanceSerializer('vtkOpenGLProperty', propertySerializer)

  # Datasets
  registerInstanceSerializer('vtkPolyData', polydataSerializer)
  registerInstanceSerializer('vtkMultiBlockDataSet', mergeToPolydataSerializer)

  # RenderWindows
  registerInstanceSerializer('vtkCocoaRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkXOpenGLRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkWin32OpenGLRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkEGLRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkOpenVRRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkGenericOpenGLRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkOSOpenGLRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkOpenGLRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkIOSRenderWindow', renderWindowSerializer)
  registerInstanceSerializer('vtkExternalOpenGLRenderWindow', renderWindowSerializer)

  # Renderers
  registerInstanceSerializer('vtkOpenGLRenderer', rendererSerializer)

  # Cameras
  registerInstanceSerializer('vtkOpenGLCamera', cameraSerializer)

  # Lights
  registerInstanceSerializer('vtkPVLight', lightSerializer)
  registerInstanceSerializer('vtkOpenGLLight', lightSerializer)

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------

def pad(depth):
  padding = ''
  for d in range(depth):
    padding += '  '
  return padding

# -----------------------------------------------------------------------------

def wrapId(idStr):
  return 'instance:${%s}' % idStr

# -----------------------------------------------------------------------------

def getReferenceId(ref):
  return vtkWebApplication.GetObjectId(ref)

# -----------------------------------------------------------------------------

dataArrayShaMapping = {}

def digest(array):
  objId = getReferenceId(array)

  record = None
  if objId in dataArrayShaMapping:
    record = dataArrayShaMapping[objId]

  if record and record['mtime'] == array.GetMTime():
    return record['sha']

  record = {
    'sha': hashDataArray(array),
    'mtime': array.GetMTime()
  }

  dataArrayShaMapping[objId] = record
  return record['sha']

# -----------------------------------------------------------------------------

def getRangeInfo(array, component):
  r = array.GetRange(component)
  compRange = {}
  compRange['min'] = r[0]
  compRange['max'] = r[1]
  compRange['component'] = array.GetComponentName(component)
  return compRange

# -----------------------------------------------------------------------------

def getArrayDescription(array, context):
  if not array:
    return None

  pMd5 = digest(array)
  context.cacheDataArray(pMd5, {
    'array': array,
    'mTime': array.GetMTime(),
    'ts': time.time()
  })

  root = {}
  root['hash'] = pMd5
  root['vtkClass'] = 'vtkDataArray'
  root['name'] = array.GetName()
  root['dataType'] = getJSArrayType(array)
  root['numberOfComponents'] = array.GetNumberOfComponents()
  root['size'] = array.GetNumberOfComponents() * array.GetNumberOfTuples()
  root['ranges'] = []
  if root['numberOfComponents'] > 1:
    for i in range(root['numberOfComponents']):
      root['ranges'].append(getRangeInfo(array, i))
    root['ranges'].append(getRangeInfo(array, -1))
  else:
    root['ranges'].append(getRangeInfo(array, 0))

  return root

# -----------------------------------------------------------------------------

def extractRequiredFields(extractedFields, mapper, dataset, context, requestedFields=['Normals', 'TCoords']):
  # FIXME should evolve and support funky mapper which leverage many arrays
  if mapper.IsA('vtkMapper'):
    scalarVisibility = mapper.GetScalarVisibility()
    arrayAccessMode = mapper.GetArrayAccessMode()
    colorArrayName = mapper.GetArrayName() if arrayAccessMode == 1 else mapper.GetArrayId()
    colorMode = mapper.GetColorMode()
    scalarMode = mapper.GetScalarMode()
    if scalarVisibility and scalarMode == 3:
      arrayMeta = getArrayDescription(dataset.GetPointData().GetArray(colorArrayName), context)
      if arrayMeta:
        arrayMeta['location'] = 'pointData';
        extractedFields.append(arrayMeta)
    if scalarVisibility and scalarMode == 4:
      arrayMeta = getArrayDescription(dataset.GetCellData().GetArray(colorArrayName), context)
      if arrayMeta:
        arrayMeta['location'] = 'cellData';
        extractedFields.append(arrayMeta)

  # Normal handling
  if 'Normals' in requestedFields:
    normals = dataset.GetPointData().GetNormals()
    if normals:
      arrayMeta = getArrayDescription(normals, context)
      if arrayMeta:
        arrayMeta['location'] = 'pointData'
        arrayMeta['registration'] = 'setNormals'
        extractedFields.append(arrayMeta)

  # TCoord handling
  if 'TCoords' in requestedFields:
    tcoords = dataset.GetPointData().GetTCoords()
    if tcoords:
      arrayMeta = getArrayDescription(tcoords, context)
      if arrayMeta:
        arrayMeta['location'] = 'pointData'
        arrayMeta['registration'] = 'setTCoords'
        extractedFields.append(arrayMeta)

# -----------------------------------------------------------------------------
# Concrete instance serializers
# -----------------------------------------------------------------------------

def genericActorSerializer(parent, actor, actorId, context, depth):
  # This kind of actor has two "children" of interest, a property and a mapper
  actorVisibility = actor.GetVisibility()
  mapperInstance = None
  propertyInstance = None
  calls = []
  dependencies = []

  if actorVisibility:
    mapper = None
    if not hasattr(actor, 'GetMapper'):
      if context.debugAll: print('This actor does not have a GetMapper method')
    else:
      mapper = actor.GetMapper()

    if mapper:
      mapperId = getReferenceId(mapper)
      mapperInstance = serializeInstance(actor, mapper, mapperId, context, depth + 1)
      if mapperInstance:
        dependencies.append(mapperInstance)
        calls.append(['setMapper', [ wrapId(mapperId) ]])

    prop = None
    if hasattr(actor, 'GetProperty'):
      prop = actor.GetProperty()
    else:
      if context.debugAll: print('This actor does not have a GetProperty method')

    if prop:
      propId = getReferenceId(prop)
      propertyInstance = serializeInstance(actor, prop, propId, context, depth + 1)
      if propertyInstance:
        dependencies.append(propertyInstance)
        calls.append(['setProperty', [ wrapId(propId) ]])

  if actorVisibility == 0 or (mapperInstance and propertyInstance):
    return {
      'parent': getReferenceId(parent),
      'id': actorId,
      'type': actor.GetClassName(),
      'properties': {
        # vtkProp
        'visibility': actorVisibility,
        'pickable': actor.GetPickable(),
        'dragable': actor.GetDragable(),
        'useBounds': actor.GetUseBounds(),
        # vtkProp3D
        'origin': actor.GetOrigin(),
        'position': actor.GetPosition(),
        'scale': actor.GetScale(),
        # vtkActor
        'forceOpaque': actor.GetForceOpaque(),
        'forceTranslucent': actor.GetForceTranslucent()
      },
      'calls': calls,
      'dependencies': dependencies
    }

  return None

# -----------------------------------------------------------------------------

def genericMapperSerializer(parent, mapper, mapperId, context, depth):
  # This kind of mapper requires us to get 2 items: input data and lookup table
  dataObject = None
  dataObjectInstance = None
  lookupTableInstance = None
  calls = []
  dependencies = []

  if hasattr(mapper, 'GetInputDataObject'):
    dataObject = mapper.GetInputDataObject(0, 0)
  else:
    if context.debugAll: print('This mapper does not have GetInputDataObject method')

  if dataObject:
    dataObjectId = '%s-dataset' % mapperId
    dataObjectInstance = serializeInstance(mapper, dataObject, dataObjectId, context, depth + 1)
    if dataObjectInstance:
      dependencies.append(dataObjectInstance)
      calls.append(['setInputData', [ wrapId(dataObjectId) ]])

  lookupTable = None

  if hasattr(mapper, 'GetLookupTable'):
    lookupTable = mapper.GetLookupTable()
  else:
    if context.debugAll: print('This mapper does not have GetLookupTable method')

  if lookupTable:
    lookupTableId = getReferenceId(lookupTable)
    lookupTableInstance = serializeInstance(mapper, lookupTable, lookupTableId, context, depth + 1)
    if lookupTableInstance:
      dependencies.append(lookupTableInstance)
      calls.append(['setLookupTable', [ wrapId(lookupTableId) ]])

  if dataObjectInstance and lookupTableInstance:
    colorArrayName = mapper.GetArrayName() if mapper.GetArrayAccessMode() == 1 else mapper.GetArrayId()
    return {
      'parent': getReferenceId(parent),
      'id': mapperId,
      'type': mapper.GetClassName(),
      'properties': {
        'scalarRange': mapper.GetScalarRange(),
        'useLookupTableScalarRange': True if mapper.GetUseLookupTableScalarRange() else False,
        'scalarVisibility': mapper.GetScalarVisibility(),
        'colorByArrayName': colorArrayName,
        'colorMode': mapper.GetColorMode(),
        'scalarMode': mapper.GetScalarMode(),
        'interpolateScalarsBeforeMapping': True if mapper.GetInterpolateScalarsBeforeMapping() else False
      },
      'calls': calls,
      'dependencies': dependencies
    }

  return None

# -----------------------------------------------------------------------------

def lookupTableSerializer(parent, lookupTable, lookupTableId, context, depth):
  # No children in this case, so no additions to bindings and return empty list
  # But we do need to add instance

  lookupTableRange = lookupTable.GetRange()

  lookupTableHueRange = [0.5, 0]
  if hasattr(lookupTable, 'GetHueRange'):
    try:
      lookupTable.GetHueRange(lookupTableHueRange)
    except Exception as inst:
      pass

  lutSatRange = lookupTable.GetSaturationRange()
  lutAlphaRange = lookupTable.GetAlphaRange()

  return {
    'parent': getReferenceId(parent),
    'id': lookupTableId,
    'type': lookupTable.GetClassName(),
    'properties': {
      'numberOfColors': lookupTable.GetNumberOfColors(),
      'valueRange': lookupTableRange,
      'hueRange': lookupTableHueRange,
      # 'alphaRange': lutAlphaRange,  # Causes weird rendering artifacts on client
      'saturationRange': lutSatRange,
      'nanColor': lookupTable.GetNanColor(),
      'belowRangeColor': lookupTable.GetBelowRangeColor(),
      'aboveRangeColor': lookupTable.GetAboveRangeColor(),
      'useAboveRangeColor': True if lookupTable.GetUseAboveRangeColor() else False,
      'useBelowRangeColor': True if lookupTable.GetUseBelowRangeColor() else False,
      'alpha': lookupTable.GetAlpha(),
      'vectorSize': lookupTable.GetVectorSize(),
      'vectorComponent': lookupTable.GetVectorComponent(),
      'vectorMode': lookupTable.GetVectorMode(),
      'indexedLookup': lookupTable.GetIndexedLookup()
    }
  }

# -----------------------------------------------------------------------------

def propertySerializer(parent, propObj, propObjId, context, depth):
  representation = propObj.GetRepresentation() if hasattr(propObj, 'GetRepresentation') else 2
  colorToUse = propObj.GetDiffuseColor() if hasattr(propObj, 'GetDiffuseColor') else [1, 1, 1]
  if representation == 1 and hasattr(propObj, 'GetColor'):
    colorToUse = propObj.GetColor()

  return {
    'parent': getReferenceId(parent),
    'id': propObjId,
    'type': propObj.GetClassName(),
    'properties': {
      'representation': representation,
      'diffuseColor': colorToUse,
      'color': propObj.GetColor(),
      'ambientColor': propObj.GetAmbientColor(),
      'specularColor': propObj.GetSpecularColor(),
      'edgeColor': propObj.GetEdgeColor(),
      'ambient': propObj.GetAmbient(),
      'diffuse': propObj.GetDiffuse(),
      'specular': propObj.GetSpecular(),
      'specularPower': propObj.GetSpecularPower(),
      'opacity': propObj.GetOpacity(),
      'interpolation': propObj.GetInterpolation(),
      'edgeVisibility': True if propObj.GetEdgeVisibility() else False,
      'backfaceCulling': True if propObj.GetBackfaceCulling() else False,
      'frontfaceCulling': True if propObj.GetFrontfaceCulling() else False,
      'pointSize': propObj.GetPointSize(),
      'lineWidth': propObj.GetLineWidth(),
      'lighting': propObj.GetLighting()
    }
  }

# -----------------------------------------------------------------------------

def polydataSerializer(parent, dataset, datasetId, context, depth):
  datasetType = dataset.GetClassName()

  if dataset and dataset.GetPoints():
    properties = {}

    # Points
    points = getArrayDescription(dataset.GetPoints().GetData(), context)
    points['vtkClass'] = 'vtkPoints'
    properties['points'] = points

    ## Verts
    if dataset.GetVerts() and dataset.GetVerts().GetData().GetNumberOfTuples() > 0:
      _verts = getArrayDescription(dataset.GetVerts().GetData(), context)
      properties['verts'] = _verts
      properties['verts']['vtkClass'] = 'vtkCellArray'

    ## Lines
    if dataset.GetLines() and dataset.GetLines().GetData().GetNumberOfTuples() > 0:
      _lines = getArrayDescription(dataset.GetLines().GetData(), context)
      properties['lines'] = _lines
      properties['lines']['vtkClass'] = 'vtkCellArray'

    ## Polys
    if dataset.GetPolys() and dataset.GetPolys().GetData().GetNumberOfTuples() > 0:
      _polys = getArrayDescription(dataset.GetPolys().GetData(), context)
      properties['polys'] = _polys
      properties['polys']['vtkClass'] = 'vtkCellArray'

    ## Strips
    if dataset.GetStrips() and dataset.GetStrips().GetData().GetNumberOfTuples() > 0:
      _strips = getArrayDescription(dataset.GetStrips().GetData(), context)
      properties['strips'] = _strips
      properties['strips']['vtkClass'] = 'vtkCellArray'

    ## Fields
    properties['fields'] = []
    extractRequiredFields(properties['fields'], parent, dataset, context)

    return {
      'parent': getReferenceId(parent),
      'id': datasetId,
      'type': datasetType,
      'properties': properties
    }

  if context.debugAll: print('This dataset has no points!')
  return None

# -----------------------------------------------------------------------------

def mergeToPolydataSerializer(parent, dataObject, dataObjectId, context, depth):
  dataset = None

  if dataObject.IsA('vtkCompositeDataSet'):
    dataMTime = dataObject.GetMTime()
    gf = vtkCompositeDataGeometryFilter()
    gf.SetInputData(dataObject)
    gf.Update()
    tempDS = gf.GetOutput()
    dataset = tempDS
  else:
    dataset = mapper.GetInput()

  return polydataSerializer(parent, dataset, dataObjectId, context, depth)

# -----------------------------------------------------------------------------

def colorTransferFunctionSerializer(parent, instance, objId, context, depth):
  nodes = []

  for i in range(instance.GetSize()):
    # x, r, g, b, midpoint, sharpness
    node = [0, 0, 0, 0, 0, 0]
    instance.GetNodeValue(i, node)
    nodes.append(node)

  return {
    'parent': getReferenceId(parent),
    'id': objId,
    'type': instance.GetClassName(),
    'properties': {
      'clamping': True if instance.GetClamping() else False,
      'colorSpace': instance.GetColorSpace(),
      'hSVWrap': True if instance.GetHSVWrap() else False,
      # 'nanColor': instance.GetNanColor(),                  # Breaks client
      # 'belowRangeColor': instance.GetBelowRangeColor(),    # Breaks client
      # 'aboveRangeColor': instance.GetAboveRangeColor(),    # Breaks client
      # 'useAboveRangeColor': True if instance.GetUseAboveRangeColor() else False,
      # 'useBelowRangeColor': True if instance.GetUseBelowRangeColor() else False,
      'allowDuplicateScalars': True if instance.GetAllowDuplicateScalars() else False,
      'alpha': instance.GetAlpha(),
      'vectorComponent': instance.GetVectorComponent(),
      'vectorSize': instance.GetVectorSize(),
      'vectorMode': instance.GetVectorMode(),
      'indexedLookup': instance.GetIndexedLookup(),
      'nodes': nodes
    }
  }

# -----------------------------------------------------------------------------

def rendererSerializer(parent, instance, objId, context, depth):
  dependencies = []
  viewPropIds = []
  lightsIds = []
  calls = []

  # Camera
  camera = instance.GetActiveCamera()
  cameraId = getReferenceId(camera)
  cameraBindings = []
  cameraInstance = serializeInstance(instance, camera, cameraId, context, depth + 1)
  if cameraInstance:
    dependencies.append(cameraInstance)
    calls.append(['setActiveCamera', [ wrapId(cameraId) ]])

  # View prop as representation containers
  viewPropCollection = instance.GetViewProps()
  for rpIdx in range(viewPropCollection.GetNumberOfItems()):
    viewProp = viewPropCollection.GetItemAsObject(rpIdx)
    viewPropId = getReferenceId(viewProp)

    viewPropInstance = serializeInstance(instance, viewProp, viewPropId, context, depth + 1)
    if viewPropInstance:
      dependencies.append(viewPropInstance)
      viewPropIds.append(viewPropId)

  calls += context.buildDependencyCallList('%s-props' % objId, viewPropIds, 'addViewProp', 'removeViewProp')

  # Lights
  lightCollection = instance.GetLights()
  for lightIdx in range(lightCollection.GetNumberOfItems()):
    light = lightCollection.GetItemAsObject(lightIdx)
    lightId = getReferenceId(light)

    lightInstance = serializeInstance(instance, light, lightId, context, depth + 1)
    if lightInstance:
      dependencies.append(lightInstance)
      lightsIds.append(lightId)

  calls += context.buildDependencyCallList('%s-lights' % objId, lightsIds, 'addLight', 'removeLight')

  if len(dependencies) > 1:
    return {
      'parent': getReferenceId(parent),
      'id': objId,
      'type': instance.GetClassName(),
      'properties': {
        'background': instance.GetBackground(),
        'background2': instance.GetBackground2(),
        'viewport': instance.GetViewport(),
        ### These commented properties do not yet have real setters in vtk.js
        # 'gradientBackground': instance.GetGradientBackground(),
        # 'aspect': instance.GetAspect(),
        # 'pixelAspect': instance.GetPixelAspect(),
        # 'ambient': instance.GetAmbient(),
        'twoSidedLighting': instance.GetTwoSidedLighting(),
        'lightFollowCamera': instance.GetLightFollowCamera(),
        'layer': instance.GetLayer(),
        'preserveColorBuffer': instance.GetPreserveColorBuffer(),
        'preserveDepthBuffer': instance.GetPreserveDepthBuffer(),
        'nearClippingPlaneTolerance': instance.GetNearClippingPlaneTolerance(),
        'clippingRangeExpansion': instance.GetClippingRangeExpansion(),
        'useShadows': instance.GetUseShadows(),
        'useDepthPeeling': instance.GetUseDepthPeeling(),
        'occlusionRatio': instance.GetOcclusionRatio(),
        'maximumNumberOfPeels': instance.GetMaximumNumberOfPeels()
      },
      'dependencies': dependencies,
      'calls': calls
    }

  return None

# -----------------------------------------------------------------------------

def cameraSerializer(parent, instance, objId, context, depth):
  return {
    'parent': getReferenceId(parent),
    'id': objId,
    'type': instance.GetClassName(),
    'properties': {
      'focalPoint': instance.GetFocalPoint(),
      'position': instance.GetPosition(),
      'viewUp': instance.GetViewUp(),
    }
  }

# -----------------------------------------------------------------------------

def lightTypeToString(value):
  """
  #define VTK_LIGHT_TYPE_HEADLIGHT    1
  #define VTK_LIGHT_TYPE_CAMERA_LIGHT 2
  #define VTK_LIGHT_TYPE_SCENE_LIGHT  3

  'HeadLight';
  'SceneLight';
  'CameraLight'
  """
  if value == 1:
    return 'HeadLight'
  elif value == 2:
    return 'CameraLight'

  return 'SceneLight'

def lightSerializer(parent, instance, objId, context, depth):
  return {
    'parent': getReferenceId(parent),
    'id': objId,
    'type': instance.GetClassName(),
    'properties': {
      # 'specularColor': instance.GetSpecularColor(),
      # 'ambientColor': instance.GetAmbientColor(),
      'switch': instance.GetSwitch(),
      'intensity': instance.GetIntensity(),
      'color': instance.GetDiffuseColor(),
      'position': instance.GetPosition(),
      'focalPoint': instance.GetFocalPoint(),
      'positional': instance.GetPositional(),
      'exponent': instance.GetExponent(),
      'coneAngle': instance.GetConeAngle(),
      'attenuationValues': instance.GetAttenuationValues(),
      'lightType': lightTypeToString(instance.GetLightType()),
      'shadowAttenuation': instance.GetShadowAttenuation()
    }
  }

# -----------------------------------------------------------------------------

def renderWindowSerializer(parent, instance, objId, context, depth):
  dependencies = []
  rendererIds = []

  rendererCollection = instance.GetRenderers()
  for rIdx in range(rendererCollection.GetNumberOfItems()):
    # Grab the next vtkRenderer
    renderer = rendererCollection.GetItemAsObject(rIdx)
    rendererId = getReferenceId(renderer)
    rendererInstance = serializeInstance(instance, renderer, rendererId, context, depth + 1)
    if rendererInstance:
      dependencies.append(rendererInstance)
      rendererIds.append(rendererId)

  calls = context.buildDependencyCallList(objId, rendererIds, 'addRenderer', 'removeRenderer')

  return {
    'parent': getReferenceId(parent),
    'id': objId,
    'type': instance.GetClassName(),
    'properties': {
      'numberOfLayers': instance.GetNumberOfLayers()
    },
    'dependencies': dependencies,
    'calls': calls
  }



from vtk import *
from math import *

# -----------------------------------------------------------------------------
# Set of helper functions
# -----------------------------------------------------------------------------

def normalize(vect, tolerance=0.00001):
    mag2 = sum(n * n for n in vect)
    if abs(mag2 - 1.0) > tolerance:
        mag = sqrt(mag2)
        vect = tuple(n / mag for n in vect)
    return vect

def q_mult(q1, q2):
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return w, x, y, z

def q_conjugate(q):
    w, x, y, z = q
    return (w, -x, -y, -z)

def qv_mult(q1, v1):
    q2 = (0.0,) + v1
    return q_mult(q_mult(q1, q2), q_conjugate(q1))[1:]

def axisangle_to_q(v, theta):
    v = normalize(v)
    x, y, z = v
    theta /= 2
    w = cos(theta)
    x = x * sin(theta)
    y = y * sin(theta)
    z = z * sin(theta)
    return w, x, y, z

def vectProduct(axisA, axisB):
    xa, ya, za = axisA
    xb, yb, zb = axisB
    normalVect = (ya*zb - za*yb, za*xb - xa*zb, xa*yb - ya*xb)
    normalVect = normalize(normalVect)
    return normalVect

def dotProduct(vecA, vecB):
    return (vecA[0] * vecB[0]) + (vecA[1] * vecB[1]) + (vecA[2] * vecB[2])

def rotate(axis, angle, center, point):
    angleInRad = 3.141592654 * angle / 180.0
    rotation = axisangle_to_q(axis, angleInRad)
    tPoint = tuple((point[i] - center[i]) for i in range(3))
    rtPoint = qv_mult(rotation, tPoint)
    rPoint = tuple((rtPoint[i] + center[i]) for i in range(3))
    return rPoint

# -----------------------------------------------------------------------------
# Spherical Camera
# -----------------------------------------------------------------------------

class SphericalCamera(object):

    def __init__(self, dataHandler, focalPoint, position, phiAxis, phiAngles, thetaAngles):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.thetaBind = { "mouse" : { "drag" : { "modifier": 0, "coordinate": 1, "step": 30 , "orientation": 1} } }
        self.phiBind = { "mouse" : { "drag" : { "modifier": 0, "coordinate": 0, "step": 30 , "orientation": 1} } }

        # Convert to serializable type
        fp = tuple(i for i in focalPoint)

        # Register arguments to the data handler
        if len(phiAngles) > 1 and phiAngles[-1] + phiAngles[1] == 360:
            self.dataHandler.registerArgument(priority=0, name='phi', values=phiAngles, ui='slider', loop='modulo', bind=self.phiBind)
        else:
            self.dataHandler.registerArgument(priority=0, name='phi', values=phiAngles, ui='slider', bind=self.phiBind)
        if thetaAngles[0] < 0 and thetaAngles[0] >= -90:
            idx = 0
            for theta in thetaAngles:
                if theta < 0:
                    idx += 1

            self.dataHandler.registerArgument(priority=0, name='theta', values=[ (x+90) for x in thetaAngles ], ui='slider', default=idx, bind=self.thetaBind)
        else:
            self.dataHandler.registerArgument(priority=0, name='theta', values=thetaAngles, ui='slider', bind=self.thetaBind)

        # Compute all camera settings
        for theta in thetaAngles:
            for phi in phiAngles:
                phiPos = rotate(phiAxis, -phi, fp, position)
                thetaAxis = vectProduct(phiAxis, tuple(fp[i]-phiPos[i] for i in range(3)))
                thetaPhiPos = rotate(thetaAxis, theta, fp, phiPos)
                viewUp = rotate(thetaAxis, theta, (0,0,0), phiAxis)

                self.cameraSettings.append({
                    'theta': theta,
                    'thetaIdx': thetaAngles.index(theta),
                    'phi': phi,
                    'phiIdx': phiAngles.index(phi),
                    'focalPoint': fp,
                    'position': thetaPhiPos,
                    'viewUp': viewUp
                })

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ['theta', 'phi']
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for cameraData in self.cameraSettings:
            self.dataHandler.setArguments(phi=cameraData['phiIdx'], theta=cameraData['thetaIdx'])
            yield cameraData

# -----------------------------------------------------------------------------
# Cylindrical Camera
# -----------------------------------------------------------------------------

class CylindricalCamera(object):

    def __init__(self, dataHandler, focalPoint, position, rotationAxis, phiAngles, translationValues):
        self.dataHandler = dataHandler
        self.cameraSettings = []

        # Register arguments to the data handler
        self.dataHandler.registerArgument(priority=0, name='phi', values=phiAngles, ui='slider', loop='modulo')
        self.dataHandler.registerArgument(priority=0, name='n_pos', values=translationValues, ui='slider')

        # Compute all camera settings
        for translation in translationValues:
            for phi in phiAngles:
                phiPos = rotate(rotationAxis, phi, focalPoint, position)
                newfocalPoint = tuple(focalPoint[i] + (translation*rotationAxis[i]) for i in range(3))
                transPhiPoint = tuple(phiPos[i] + (translation*rotationAxis[i]) for i in range(3))

                self.cameraSettings.append({
                    'n_pos': translation,
                    'n_posIdx': translationValues.index(translation),
                    'phi': phi,
                    'phiIdx': phiAngles.index(phi),
                    'focalPoint': newfocalPoint,
                    'position': transPhiPoint,
                    'viewUp': rotationAxis
                })

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ['n_pos', 'phi']
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for cameraData in self.cameraSettings:
            self.dataHandler.setArguments(phi=cameraData['phiIdx'], n_pos=cameraData['n_posIdx'])
            yield cameraData

# -----------------------------------------------------------------------------
# MultiView Cube Camera
# -----------------------------------------------------------------------------

class CubeCamera(object):

    # positions = [ { position: [x,y,z], args: { i: 1, j: 0, k: 7 } }, ... ]
    def __init__(self, dataHandler, viewForward, viewUp, positions):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.viewForward = viewForward
        self.viewUp = viewUp
        self.rightDirection = vectProduct(viewForward, viewUp)
        self.positions = positions

        # Register arguments to the data handler
        self.dataHandler.registerArgument(priority=0, name='orientation', values=['f', 'b', 'r', 'l', 'u', 'd'])

        # Register arguments to id position
        self.args = {}
        for pos in positions:
            for key in pos['args']:
                if key not in self.args:
                    self.args[key] = {}
                self.args[key][pos['args'][key]] = True

        for key in self.args:
            self.args[key] = sorted(self.args[key], key=lambda k: int(k))

        self.keyList = self.args.keys()
        for key in self.args:
            self.dataHandler.registerArgument(priority=1, name=key, values=self.args[key])

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ['orientation']
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for pos in self.positions:
            cameraData = {
                'position': pos['position'],
            }

            print('='*80)
            for key in pos['args']:
                idx = self.args[key].index(pos['args'][key])
                self.dataHandler.setArguments(**{ key: idx })
                print(key, idx)

            print('position', cameraData['position'])

            # front
            cameraData['focalPoint'] = [(cameraData['position'][i] + self.viewForward[i]) for i in range(3)]
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['orientation'] = 'front'
            self.dataHandler.setArguments(orientation=0)
            yield cameraData

            # back
            cameraData['focalPoint'] = [(cameraData['position'][i] - self.viewForward[i]) for i in range(3)]
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['orientation'] = 'back'
            self.dataHandler.setArguments(orientation=1)
            yield cameraData

            # right
            self.dataHandler.setArguments(orientation=2)
            cameraData['focalPoint'] = [(cameraData['position'][i] + self.rightDirection[i]) for i in range(3)]
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['orientation'] = 'right'
            yield cameraData

            # left
            self.dataHandler.setArguments(orientation=3)
            cameraData['focalPoint'] = [(cameraData['position'][i] - self.rightDirection[i]) for i in range(3)]
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['orientation'] = 'left'
            yield cameraData

            # up
            self.dataHandler.setArguments(orientation=4)
            cameraData['focalPoint'] = [(cameraData['position'][i] + self.viewUp[i]) for i in range(3)]
            cameraData['viewUp'] = [(-self.viewForward[i]) for i in range(3)]
            cameraData['orientation'] = 'up'
            yield cameraData

            # doww
            self.dataHandler.setArguments(orientation=5)
            cameraData['focalPoint'] = [(cameraData['position'][i] - self.viewUp[i]) for i in range(3)]
            cameraData['viewUp'] = [self.viewForward[i] for i in range(3)]
            cameraData['orientation'] = 'down'
            yield cameraData

# -----------------------------------------------------------------------------
# MultiView Cube Camera
# -----------------------------------------------------------------------------

class StereoCubeCamera(object):

    # positions = [ { position: [x,y,z], args: { i: 1, j: 0, k: 7 } }, ... ]
    def __init__(self, dataHandler, viewForward, viewUp, positions, eyeSpacing):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.viewForward = viewForward
        self.viewUp = viewUp
        self.rightDirection = vectProduct(viewForward, viewUp)
        self.positions = positions
        self.eyeSpacing = eyeSpacing

        # Register arguments to the data handler
        self.dataHandler.registerArgument(priority=0, name='orientation', values=['f', 'b', 'r', 'l', 'u', 'd'])
        self.dataHandler.registerArgument(priority=0, name='eye', values=['left', 'right'])

        # Register arguments to id position
        self.args = {}
        for pos in positions:
            for key in pos['args']:
                if key not in self.args:
                    self.args[key] = {}
                self.args[key][pos['args'][key]] = True

        for key in self.args:
            self.args[key] = sorted(self.args[key], key=lambda k: int(k))

        self.keyList = self.args.keys()
        for key in self.args:
            self.dataHandler.registerArgument(priority=1, name=key, values=self.args[key])

        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ['orientation']
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for pos in self.positions:
            cameraData = {}

            for key in pos['args']:
                idx = self.args[key].index(pos['args'][key])
                self.dataHandler.setArguments(**{ key: idx })

            # front
            cameraData['orientation'] = 'front'
            self.dataHandler.setArguments(orientation=0)
            deltaVect = [(v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] - deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] + self.viewForward[i] - deltaVect[i]) for i in range(3)]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] + deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] + self.viewForward[i] + deltaVect[i]) for i in range(3)]
            yield cameraData

            # back
            cameraData['orientation'] = 'back'
            self.dataHandler.setArguments(orientation=1)
            deltaVect = [-(v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] - deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] - self.viewForward[i] - deltaVect[i]) for i in range(3)]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] + deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] - self.viewForward[i] + deltaVect[i]) for i in range(3)]
            yield cameraData

            # right
            self.dataHandler.setArguments(orientation=2)
            cameraData['orientation'] = 'right'
            deltaVect = [-(v * float(self.eyeSpacing) * 0.5) for v in self.viewForward]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] - deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] + self.rightDirection[i] - deltaVect[i]) for i in range(3)]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] + deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] + self.rightDirection[i] + deltaVect[i]) for i in range(3)]
            yield cameraData

            # left
            self.dataHandler.setArguments(orientation=3)
            cameraData['orientation'] = 'left'
            deltaVect = [(v * float(self.eyeSpacing) * 0.5) for v in self.viewForward]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] - deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] - self.rightDirection[i] - deltaVect[i]) for i in range(3)]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData['viewUp'] = [self.viewUp[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] + deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] - self.rightDirection[i] + deltaVect[i]) for i in range(3)]
            yield cameraData

            # up
            self.dataHandler.setArguments(orientation=4)
            cameraData['orientation'] = 'up'
            deltaVect = [(v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData['viewUp'] = [(-self.viewForward[i]) for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] - deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] + self.viewUp[i] - deltaVect[i]) for i in range(3)]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData['viewUp'] = [(-self.viewForward[i]) for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] + deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] + self.viewUp[i] + deltaVect[i]) for i in range(3)]
            yield cameraData

            # doww
            self.dataHandler.setArguments(orientation=5)
            cameraData['orientation'] = 'down'
            deltaVect = [(v * float(self.eyeSpacing) * 0.5) for v in self.rightDirection]
            ## Left-Eye
            self.dataHandler.setArguments(eye=0)
            cameraData['viewUp'] = [self.viewForward[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] - deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] - self.viewUp[i] - deltaVect[i]) for i in range(3)]
            yield cameraData
            ## Right-Eye
            self.dataHandler.setArguments(eye=1)
            cameraData['viewUp'] = [self.viewForward[i] for i in range(3)]
            cameraData['position'] = [(pos['position'][idx] + deltaVect[idx]) for idx in range(3)]
            cameraData['focalPoint'] = [(pos['position'][i] - self.viewUp[i] + deltaVect[i]) for i in range(3)]
            yield cameraData

# -----------------------------------------------------------------------------
# MultiView Camera
# -----------------------------------------------------------------------------

class MultiViewCamera(object):

    def __init__(self, dataHandler):
        self.dataHandler = dataHandler
        self.cameraSettings = []
        self.positionNames = []

    def registerViewPoint(self, name, focalPoint, position, viewUp):
        self.cameraSettings.append({'name': name, 'nameIdx': len(self.positionNames), 'focalPoint': focalPoint, 'position': position, 'viewUp': viewUp})
        self.positionNames.append(name)
        self.dataHandler.registerArgument(priority=0, name='multiView', values=self.positionNames)
        self.dataHandler.updateBasePattern()

    def updatePriority(self, priorityList):
        keyList = ['multiView']
        for idx in range(min(len(priorityList), len(keyList))):
            self.dataHandler.updatePriority(keyList[idx], priorityList[idx])

    def __iter__(self):
        for cameraData in self.cameraSettings:
            self.dataHandler.setArguments(multiView=cameraData['nameIdx'])
            yield cameraData

# -----------------------------------------------------------------------------
# Helper methods
# -----------------------------------------------------------------------------

def update_camera(renderer, cameraData):
    camera = renderer.GetActiveCamera()
    camera.SetPosition(cameraData['position'])
    camera.SetFocalPoint(cameraData['focalPoint'])
    camera.SetViewUp(cameraData['viewUp'])

def create_spherical_camera(renderer, dataHandler, phiValues, thetaValues):
    camera = renderer.GetActiveCamera()
    return SphericalCamera(dataHandler, camera.GetFocalPoint(), camera.GetPosition(), camera.GetViewUp(), phiValues, thetaValues)

def create_cylindrical_camera(renderer, dataHandler, phiValues, translationValues):
    camera = renderer.GetActiveCamera()
    return CylindricalCamera(dataHandler, camera.GetFocalPoint(), camera.GetPosition(), camera.GetViewUp(), phiValues, translationValues)



from vtkmodules.web import camera
from paraview.web import data_writer
from paraview.web import data_converter

from vtkmodules.web.query_data_model import *
from paraview.web.camera import *
from vtkmodules.web import iteritems, buffer
from paraview import simple
from paraview import servermanager

from vtkmodules.vtkCommonCore import vtkFloatArray, vtkIdList, vtkUnsignedCharArray, vtkTypeUInt32Array
from vtkmodules.vtkCommonDataModel import vtkDataSetAttributes

import json, os, math, gzip, shutil, hashlib

# Global helper variables
encode_codes = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
arrayTypesMapping = '  bBhHiIlLfd'
jsMapping = {
    'b': 'Int8Array',
    'B': 'Uint8Array',
    'h': 'Int16Array',
    'H': 'Int16Array',
    'i': 'Int32Array',
    'I': 'Uint32Array',
    'l': 'Int32Array',
    'L': 'Uint32Array',
    'f': 'Float32Array',
    'd': 'Float64Array'
}

# -----------------------------------------------------------------------------
# Basic Dataset Builder
# -----------------------------------------------------------------------------

class DataSetBuilder(object):
    def __init__(self, location, camera_data, metadata={}, sections={}):
        self.dataHandler = DataHandler(location)
        self.cameraDescription = camera_data
        self.camera = None

        for key, value in iteritems(metadata):
            self.dataHandler.addMetaData(key, value)

        for key, value in iteritems(sections):
            self.dataHandler.addSection(key, value)

        # Update the can_write flag for MPI
        self.dataHandler.can_write = (servermanager.vtkProcessModule.GetProcessModule().GetPartitionId() == 0)

    def getDataHandler(self):
        return self.dataHandler

    def getCamera(self):
        return self.camera

    def updateCamera(self, camera):
        update_camera(self.view, camera)

    def start(self, view=None):
        if view:
            # Keep track of the view
            self.view = view

            # Handle camera if any
            if self.cameraDescription:
                if self.cameraDescription['type'] == 'spherical':
                    self.camera = camera.SphericalCamera(self.dataHandler, view.CenterOfRotation, view.CameraPosition, view.CameraViewUp, self.cameraDescription['phi'], self.cameraDescription['theta'])
                elif self.cameraDescription['type'] == 'cylindrical':
                    self.camera = camera.CylindricalCamera(self.dataHandler, view.CenterOfRotation, view.CameraPosition, view.CameraViewUp, self.cameraDescription['phi'], self.cameraDescription['translation'])
                elif self.cameraDescription['type'] == 'cube':
                    self.camera = camera.CubeCamera(self.dataHandler, self.cameraDescription['viewForward'], self.cameraDescription['viewUp'], self.cameraDescription['positions'])
                elif self.cameraDescription['type'] == 'cube-stereo':
                    self.camera = camera.StereoCubeCamera(self.dataHandler, self.cameraDescription['viewForward'], self.cameraDescription['viewUp'], self.cameraDescription['positions'], self.cameraDescription['eyeSeparation'])

            # Update background color
            bgColor = view.Background
            bgColorString = 'rgb(%d, %d, %d)' % tuple(int(bgColor[i]*255) for i in range(3))

            if view.UseGradientBackground:
                bgColor2 = view.Background2
                bgColor2String = 'rgb(%d, %d, %d)' % tuple(int(bgColor2[i]*255) for i in range(3))
                self.dataHandler.addMetaData('backgroundColor', 'linear-gradient(%s,%s)' % (bgColor2String, bgColorString))
            else:
                self.dataHandler.addMetaData('backgroundColor', bgColorString)

        # Update file patterns
        self.dataHandler.updateBasePattern()

    def stop(self):
        self.dataHandler.writeDataDescriptor()

# -----------------------------------------------------------------------------
# Image Dataset Builder
# -----------------------------------------------------------------------------

class ImageDataSetBuilder(DataSetBuilder):
    def __init__(self, location, imageMimeType, cameraInfo, metadata={}):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata)
        imageExtenstion = '.' + imageMimeType.split('/')[1]
        self.dataHandler.registerData(name='image', type='blob', mimeType=imageMimeType, fileName=imageExtenstion)

    def writeImages(self):
        for cam in self.camera:
            update_camera(self.view, cam)
            simple.WriteImage(self.dataHandler.getDataAbsoluteFilePath('image'))

# -----------------------------------------------------------------------------
# Data Prober Dataset Builder
# -----------------------------------------------------------------------------
class DataProberDataSetBuilder(DataSetBuilder):
    def __init__(self, input, location, sampling_dimesions, fields_to_keep, custom_probing_bounds = None, metadata={}):
        DataSetBuilder.__init__(self, location, None, metadata)
        self.fieldsToWrite = fields_to_keep
        self.resamplerFilter = simple.ResampleToImage(Input=input)
        self.resamplerFilter.SamplingDimensions = sampling_dimesions
        if custom_probing_bounds:
            self.resamplerFilter.UseInputBounds = 0
            self.resamplerFilter.SamplingBounds = custom_probing_bounds
        else:
            self.resamplerFilter.UseInputBounds = 1

        # Register all fields
        self.dataHandler.addTypes('data-prober', 'binary')
        self.DataProber = { 'types': {}, 'dimensions': sampling_dimesions, 'ranges': {}, 'spacing': [1,1,1] }
        for field in self.fieldsToWrite:
            self.dataHandler.registerData(name=field, type='array', rootFile=True, fileName='%s.array' % field)

    def writeData(self, time=0):
        if not self.dataHandler.can_write:
            return

        self.resamplerFilter.UpdatePipeline(time)
        imageData = self.resamplerFilter.GetClientSideObject().GetOutput()
        self.DataProber['spacing'] = imageData.GetSpacing()
        arrays = imageData.GetPointData()
        maskArray = arrays.GetArray(vtkDataSetAttributes.GhostArrayName())
        for field in self.fieldsToWrite:
            array = arrays.GetArray(field)
            if array:
                if array.GetNumberOfComponents() == 1:
                    # Push NaN when no value are present instead of 0
                    for idx in range(maskArray.GetNumberOfTuples()):
                        if maskArray.GetValue(idx) == 2: # Hidden point
                            array.SetValue(idx, float('NaN'))

                    with open(self.dataHandler.getDataAbsoluteFilePath(field), 'wb') as f:
                        f.write(buffer(array))

                    self.expandRange(array)
                else:
                    magarray = array.NewInstance()
                    magarray.SetNumberOfTuples(array.GetNumberOfTuples())
                    magarray.SetName(field)

                    for idx in range(magarray.GetNumberOfTuples()):
                        if maskArray.GetValue(idx) == 2: # Hidden point
                            # Push NaN when no value are present
                            magarray.SetValue(idx, float('NaN'))
                        else:
                            entry = array.GetTuple(idx)
                            mag = self.magnitude(entry)
                            magarray.SetValue(idx,mag)

                    with open(self.dataHandler.getDataAbsoluteFilePath(field), 'wb') as f:
                        f.write(buffer(magarray))

                    self.expandRange(magarray)
            else:
                print ('No array for', field)
                print (self.resamplerFilter.GetOutput())

    def magnitude(self, tuple):
        value = 0
        for item in tuple:
            value += item * item
        value = value**0.5

        return value

    def expandRange(self, array):
        field = array.GetName()
        self.DataProber['types'][field] = jsMapping[arrayTypesMapping[array.GetDataType()]]

        if field in self.DataProber['ranges']:
            dataRange = array.GetRange()
            if dataRange[0] < self.DataProber['ranges'][field][0]:
                self.DataProber['ranges'][field][0] = dataRange[0]
            if dataRange[1] > self.DataProber['ranges'][field][1]:
                self.DataProber['ranges'][field][1] = dataRange[1]
        else:
            self.DataProber['ranges'][field] = [array.GetRange()[0], array.GetRange()[1]]

    def stop(self, compress=True):
        # Rescale spacing to have the smaller value to be 1.0
        smallerValue = min(self.DataProber['spacing'])
        if smallerValue < 1.0:
            self.DataProber['spacing'] = tuple( i / smallerValue for i in self.DataProber['spacing'])

        # Push metadata
        self.dataHandler.addSection('DataProber', self.DataProber)

        # Write metadata
        DataSetBuilder.stop(self)

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print ('Compress', root)
                for name in files:
                    if '.array' in name and '.gz' not in name:
                        with open(os.path.join(root, name), 'rb') as f_in:
                            with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))


# -----------------------------------------------------------------------------
# Float Image with Layer Dataset Builder
# -----------------------------------------------------------------------------

class LayerDataSetBuilder(DataSetBuilder):
    def __init__(self, input, location, cameraInfo, imageSize=[500,500], metadata={}):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata)
        self.dataRenderer = data_writer.ScalarRenderer(isWriter=self.dataHandler.can_write)
        self.view = self.dataRenderer.getView()
        self.view.ViewSize = imageSize
        self.floatImage = {'dimensions': imageSize, 'layers': [], 'ranges': {}}
        self.layerMap = {}
        self.input = input
        self.activeLayer = None
        self.activeField = None
        self.layerChanged = False
        self.lastTime = -1

        # Update data type
        self.dataHandler.addTypes('float-image')

    def getView(self):
        return self.view

    def setActiveLayer(self, layer, field, hasMesh=False, activeSource=None):
        if activeSource:
            self.activeSource = activeSource
        else:
            self.activeSource = self.input
        needDataRegistration = False
        if layer not in self.layerMap:
            layerObj = { 'name': layer, 'array': field, 'arrays': [ field ], 'active': True, 'type': 'Float32Array', 'hasMesh': hasMesh }
            self.layerMap[layer] = layerObj
            self.floatImage['layers'].append(layerObj)
            needDataRegistration = True

            # Register layer lighting
            self.dataHandler.registerData(name='%s__light' % layer, type='array', rootFile=True, fileName='%s__light.array' % layer, categories=[ '%s__light' % layer ])

            # Register layer mesh
            if hasMesh:
                self.dataHandler.registerData(name='%s__mesh' % layer, type='array', rootFile=True, fileName='%s__mesh.array' % layer, categories=[ '%s__mesh' % layer ])

        elif field not in self.layerMap[layer]['arrays']:
            self.layerMap[layer]['arrays'].append(field)
            needDataRegistration = True

        # Keep track of the active data
        if self.activeLayer != layer:
            self.layerChanged = True
        self.activeLayer = layer
        self.activeField = field

        if needDataRegistration:
            self.dataHandler.registerData(name='%s_%s' % (layer, field), type='array', rootFile=True, fileName='%s_%s.array' % (layer, field), categories=[ '%s_%s' % (layer, field) ])

    def writeLayerData(self, time=0):
        dataRange = [0, 1]
        self.activeSource.UpdatePipeline(time)

        if self.activeField and self.activeLayer:

            if self.layerChanged or self.lastTime != time:
                self.layerChanged = False
                self.lastTime = time

                # Capture lighting information
                for camPos in self.getCamera():
                    self.view.CameraFocalPoint = camPos['focalPoint']
                    self.view.CameraPosition = camPos['position']
                    self.view.CameraViewUp = camPos['viewUp']
                    self.dataRenderer.writeLightArray(self.dataHandler.getDataAbsoluteFilePath('%s__light'%self.activeLayer), self.activeSource)

                # Capture mesh information
                if self.layerMap[self.activeLayer]['hasMesh']:
                    for camPos in self.getCamera():
                        self.view.CameraFocalPoint = camPos['focalPoint']
                        self.view.CameraPosition = camPos['position']
                        self.view.CameraViewUp = camPos['viewUp']
                        self.dataRenderer.writeMeshArray(self.dataHandler.getDataAbsoluteFilePath('%s__mesh'%self.activeLayer), self.activeSource)

            for camPos in self.getCamera():
                self.view.CameraFocalPoint = camPos['focalPoint']
                self.view.CameraPosition = camPos['position']
                self.view.CameraViewUp = camPos['viewUp']
                dataName = ('%s_%s' % (self.activeLayer, self.activeField))
                dataRange = self.dataRenderer.writeArray(self.dataHandler.getDataAbsoluteFilePath(dataName), self.activeSource, self.activeField)

            if self.activeField not in self.floatImage['ranges']:
                self.floatImage['ranges'][self.activeField] = [ dataRange[0], dataRange[1] ]
            else:
                # Expand the ranges
                if dataRange[0] < self.floatImage['ranges'][self.activeField][0]:
                    self.floatImage['ranges'][self.activeField][0] = dataRange[0]
                if dataRange[1] > self.floatImage['ranges'][self.activeField][1]:
                    self.floatImage['ranges'][self.activeField][1] = dataRange[1]

    def start(self):
        DataSetBuilder.start(self, self.view)

    def stop(self, compress=True):
        if not self.dataHandler.can_write:
            return

        # Push metadata
        self.dataHandler.addSection('FloatImage', self.floatImage)

        # Write metadata
        DataSetBuilder.stop(self)

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print ('Compress', root)
                for name in files:
                    if '.array' in name and '.gz' not in name:
                        with open(os.path.join(root, name), 'rb') as f_in:
                            with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))


# -----------------------------------------------------------------------------
# Composite Dataset Builder
# -----------------------------------------------------------------------------

class CompositeDataSetBuilder(DataSetBuilder):
    def __init__(self, location, sceneConfig, cameraInfo, metadata={}, sections={}, view=None):
        DataSetBuilder.__init__(self, location, cameraInfo, metadata, sections)

        if view:
            self.view = view
        else:
            self.view = simple.CreateView('RenderView')

        self.view.ViewSize = sceneConfig['size']
        self.view.CenterAxesVisibility = 0
        self.view.OrientationAxesVisibility = 0
        self.view.UpdatePropertyInformation()
        if 'camera' in sceneConfig and 'CameraFocalPoint' in sceneConfig['camera']:
            self.view.CenterOfRotation = sceneConfig['camera']['CameraFocalPoint']

        # Initialize camera
        for key, value in iteritems(sceneConfig['camera']):
            self.view.GetProperty(key).SetData(value)

        # Create a representation for all scene sources
        self.config = sceneConfig
        self.representations = []
        for data in self.config['scene']:
            rep = simple.Show(data['source'], self.view)
            self.representations.append(rep)
            if 'representation' in data:
                for key in data['representation']:
                    rep.GetProperty(key).SetData(data['representation'][key])

        # Add directory path
        self.dataHandler.registerData(name='directory', rootFile=True, fileName='file.txt', categories=['trash'])

    def start(self):
        DataSetBuilder.start(self, self.view)

    def stop(self, compress=True, clean=True):
        DataSetBuilder.stop(self)

        if not self.dataHandler.can_write:
            return

        # Make the config serializable
        for item in self.config['scene']:
            del item['source']

        # Write the scene to disk
        with open(os.path.join(self.dataHandler.getBasePath(), "config.json"), 'w') as f:
            f.write(json.dumps(self.config))

        dataConverter = data_converter.ConvertCompositeDataToSortedStack(self.dataHandler.getBasePath())
        dataConverter.convert()

        # Remove tmp files
        os.remove(os.path.join(self.dataHandler.getBasePath(), "index.json"))
        os.remove(os.path.join(self.dataHandler.getBasePath(), "config.json"))

        # Composite pipeline meta description
        compositePipeline = {
            'default_pipeline': '',
            'layers': [],
            'fields': {},
            'layer_fields': {},
            'pipeline': []
        }
        rootItems = {}
        fieldNameMapping = {}

        # Clean scene in config and gather ranges
        dataRanges = {}
        layerIdx = 0
        for layer in self.config['scene']:
            # Create group node if any
            if 'parent' in layer and layer['parent'] not in rootItems:
                rootItems[layer['parent']] = { 'name': layer['parent'], 'ids': [], 'children': [] }
                compositePipeline['pipeline'].append(rootItems[layer['parent']])

            # Create layer entry
            layerCode = encode_codes[layerIdx]
            layerItem = { 'name': layer['name'], 'ids': [ layerCode ]}
            compositePipeline['layers'].append(layerCode)
            compositePipeline['layer_fields'][layerCode] = []
            compositePipeline['default_pipeline'] += layerCode

            # Register layer entry in pipeline
            if 'parent' in layer:
                rootItems[layer['parent']]['children'].append(layerItem)
                rootItems[layer['parent']]['ids'].append(layerCode)
            else:
                compositePipeline['pipeline'].append(layerItem)

            # Handle color / field
            colorByList = []
            for color in layer['colors']:
                # Find color code
                if color not in fieldNameMapping:
                    colorCode = encode_codes[len(fieldNameMapping)]
                    fieldNameMapping[color] = colorCode
                    compositePipeline['fields'][colorCode] = color
                else:
                    colorCode = fieldNameMapping[color]

                # Register color code
                compositePipeline['layer_fields'][layerCode].append(colorCode)
                if len(colorByList) == 0:
                    compositePipeline['default_pipeline'] += colorCode

                values = None
                if 'constant' in layer['colors'][color]:
                    value = layer['colors'][color]['constant']
                    values = [ value, value ]
                    colorByList.append({'name': color, 'type': 'const', 'value': value})
                elif 'range' in layer['colors'][color]:
                    values = layer['colors'][color]['range']
                    colorByList.append({'name': color, 'type': 'field'})

                if values:
                    if color not in dataRanges:
                        dataRanges[color] = values
                    else:
                        dataRanges[color][0] = min(dataRanges[color][0], values[0], values[1])
                        dataRanges[color][1] = max(dataRanges[color][1], values[0], values[1])

            layer['colorBy'] = colorByList
            del layer['colors']
            layerIdx += 1

        sortedCompositeSection = {
            'dimensions': self.config['size'],
            'pipeline': self.config['scene'],
            'ranges': dataRanges,
            'layers': len(self.config['scene']),
            'light': self.config['light']
        }
        self.dataHandler.addSection('SortedComposite', sortedCompositeSection)
        self.dataHandler.addSection('CompositePipeline', compositePipeline)
        self.dataHandler.addTypes('sorted-composite', 'multi-color-by')

        self.dataHandler.removeData('directory')
        for dataToRegister in dataConverter.listData():
            self.dataHandler.registerData(**dataToRegister)

        self.dataHandler.writeDataDescriptor()

        if clean:
          for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
              print ('Clean', root)
              for name in files:
                  if name in ['camera.json']:
                      os.remove(os.path.join(root, name))

        if compress:
            for root, dirs, files in os.walk(self.dataHandler.getBasePath()):
                print ('Compress', root)
                for name in files:
                    if ('.float32' in name or '.uint8' in name) and '.gz' not in name:
                        with open(os.path.join(root, name), 'rb') as f_in:
                            with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                shutil.copyfileobj(f_in, f_out)
                        os.remove(os.path.join(root, name))

    def writeData(self):
        composite_size = len(self.representations)

        self.view.UpdatePropertyInformation()
        self.view.Background = [0,0,0]
        imageSize = self.view.ViewSize[0] * self.view.ViewSize[1]

        # Generate the heavy data
        for camPos in self.getCamera():
            self.view.CameraFocalPoint = camPos['focalPoint']
            self.view.CameraPosition = camPos['position']
            self.view.CameraViewUp = camPos['viewUp']

            # Show all representations
            for compositeIdx in range(composite_size):
                rep = self.representations[compositeIdx]
                rep.Visibility = 1

            # Fix camera bounds
            self.view.LockBounds = 0
            simple.Render(self.view)
            self.view.LockBounds = 1

            # Update destination directory
            dest_path = os.path.dirname(self.dataHandler.getDataAbsoluteFilePath('directory'))

            # Write camera information
            if self.dataHandler.can_write:
                with open(os.path.join(dest_path, "camera.json"), 'w') as f:
                    f.write(json.dumps(camPos))

            # Hide all representations
            for compositeIdx in range(composite_size):
                rep = self.representations[compositeIdx]
                rep.Visibility = 0

            # Show only active Representation
            # Extract images for each fields
            for compositeIdx in range(composite_size):
                rep = self.representations[compositeIdx]
                if compositeIdx > 0:
                    self.representations[compositeIdx - 1].Visibility = 0
                rep.Visibility = 1

                # capture Z
                simple.Render()
                zBuffer = self.view.CaptureDepthBuffer()
                with open(os.path.join(dest_path, 'depth_%d.float32' % compositeIdx), 'wb') as f:
                    f.write(buffer(zBuffer))

                # Prevent color interference
                rep.DiffuseColor = [1,1,1]

                # Handle light
                for lightType in self.config['light']:
                    if lightType == 'intensity':
                        rep.AmbientColor  = [1,1,1]
                        rep.SpecularColor = [1,1,1]

                        self.view.StartCaptureLuminance()
                        image = self.view.CaptureWindow(1)
                        imagescalars = image.GetPointData().GetScalars()
                        self.view.StopCaptureLuminance()

                        # Extract specular information
                        specularOffset = 1 # [diffuse, specular, ?]
                        imageSize = imagescalars.GetNumberOfTuples()
                        specularComponent = vtkUnsignedCharArray()
                        specularComponent.SetNumberOfComponents(1)
                        specularComponent.SetNumberOfTuples(imageSize)

                        for idx in range(imageSize):
                            specularComponent.SetValue(idx, imagescalars.GetValue(idx * 3 + specularOffset))

                        with open(os.path.join(dest_path, 'intensity_%d.uint8' % compositeIdx), 'wb') as f:
                            f.write(buffer(specularComponent))

                        # Free memory
                        image.UnRegister(None)

                    if lightType == 'normal':
                        if rep.Representation in ['Point Gaussian', 'Points', 'Outline', 'Wireframe']:
                            uniqNormal = [(camPos['position'][i] - camPos['focalPoint'][i]) for i in range(3)]
                            tmpNormalArray = vtkFloatArray()
                            tmpNormalArray.SetNumberOfComponents(1)
                            tmpNormalArray.SetNumberOfTuples(imageSize)

                            for comp in range(3):
                                tmpNormalArray.FillComponent(0, uniqNormal[comp])
                                with open(os.path.join(dest_path, 'normal_%d_%d.float32' % (compositeIdx, comp)), 'wb') as f:
                                    f.write(buffer(tmpNormalArray))
                        else:
                            for comp in range(3):
                                # Configure view to handle POINT_DATA / CELL_DATA
                                self.view.DrawCells = 0
                                self.view.ArrayNameToDraw = 'Normals'
                                self.view.ArrayComponentToDraw = comp
                                self.view.ScalarRange = [-1.0, 1.0]
                                self.view.StartCaptureValues()
                                image = self.view.CaptureWindow(1)
                                imagescalars = image.GetPointData().GetScalars()
                                self.view.StopCaptureValues()

                                # Convert RGB => Float => Write
                                floatArray = data_converter.convertRGBArrayToFloatArray(imagescalars, [-1.0, 1.0])
                                with open(os.path.join(dest_path, 'normal_%d_%d.float32' % (compositeIdx, comp)), 'wb') as f:
                                    f.write(buffer(floatArray))

                                # Free memory
                                image.UnRegister(None)

                # Handle color by
                for fieldName, fieldConfig in iteritems(self.config['scene'][compositeIdx]['colors']):
                    if 'constant' in fieldConfig:
                        # Skip nothing to render
                        continue

                    # Configure view to handle POINT_DATA / CELL_DATA
                    if fieldConfig['location'] == 'POINT_DATA':
                        self.view.DrawCells = 0
                    else:
                        self.view.DrawCells = 1

                    self.view.ArrayNameToDraw = fieldName
                    self.view.ArrayComponentToDraw = 0
                    self.view.ScalarRange = fieldConfig['range']
                    self.view.StartCaptureValues()
                    image = self.view.CaptureWindow(1)
                    imagescalars = image.GetPointData().GetScalars()
                    self.view.StopCaptureValues()

                    floatArray = data_converter.convertRGBArrayToFloatArray(imagescalars, fieldConfig['range'])
                    with open(os.path.join(dest_path, '%d_%s.float32' % (compositeIdx, fieldName)), 'wb') as f:
                        f.write(buffer(floatArray))
                        self.dataHandler.registerData(name='%d_%s' % (compositeIdx, fieldName), fileName='/%d_%s.float32' % (compositeIdx, fieldName), type='array', categories=['%d_%s' % (compositeIdx, fieldName)])

                    # Free memory
                    image.UnRegister(None)

# -----------------------------------------------------------------------------
# VTKGeometryDataSetBuilder Dataset Builder
# -----------------------------------------------------------------------------

def writeCellArray(dataHandler, currentData, cellName, inputCellArray):
    nbValues = inputCellArray.GetNumberOfTuples()
    if nbValues == 0:
        return

    outputCells = vtkTypeUInt32Array()
    outputCells.SetNumberOfTuples(nbValues)

    for valueIdx in range(nbValues):
        outputCells.SetValue(valueIdx, inputCellArray.GetValue(valueIdx))

    iBuffer = buffer(outputCells)
    iMd5 = hashlib.md5(iBuffer).hexdigest()
    iPath = os.path.join(dataHandler.getBasePath(), 'data',"%s.Uint32Array" % iMd5)
    currentData['polys'] = 'data/%s.Uint32Array' % iMd5
    with open(iPath, 'wb') as f:
        f.write(iBuffer)

class VTKGeometryDataSetBuilder(DataSetBuilder):
    def __init__(self, location, sceneConfig, metadata={}, sections={}):
        DataSetBuilder.__init__(self, location, None, metadata, sections)

        # Update data type
        self.dataHandler.addTypes('vtk-geometry');

        # Create a representation for all scene sources
        self.config = sceneConfig

        # Processing pipeline
        self.surfaceExtract = None

        # Add directory path
        self.dataHandler.registerData(priority=0, name='scene', rootFile=True, fileName='scene.json', type='json')

        # Create directory containers
        dataPath = os.path.join(location, 'data')
        for p in [dataPath]:
            if not os.path.exists(p):
                os.makedirs(p)

        # Create metadata structure
        colorToCodeMap = {}
        parentNodes = {}
        pipelineMeta = {
            'layers': [],
            'pipeline': [],
            'layer_fields': {},
            'fields': {}
        }
        geometryMeta = {
            'ranges': {},
            'layer_map': {},
        }
        self.ranges = geometryMeta['ranges']
        for item in sceneConfig['scene']:
            # Handle layer
            layerCode = encode_codes[len(pipelineMeta['layers'])]
            pipelineMeta['layers'].append(layerCode)
            geometryMeta['layer_map'][layerCode] = item['name']

            # Handle colors
            pipelineMeta['layer_fields'][layerCode] = []
            for fieldName in item['colors']:
                colorCode = None
                if fieldName in colorToCodeMap:
                    colorCode = colorToCodeMap[fieldName]
                else:
                    colorCode = encode_codes[len(colorToCodeMap)]
                    colorToCodeMap[fieldName] = colorCode
                    geometryMeta['ranges'][fieldName] = [1, -1] # FIXME we don't know the range

                pipelineMeta['layer_fields'][layerCode].append(colorCode)
                pipelineMeta['fields'][colorCode] = fieldName

            # Handle pipeline
            if 'parent' in item:
                # Need to handle hierarchy
                if item['parent'] in parentNodes:
                    # Fill children
                    rootNode = parentNodes[item['parent']]
                    rootNode['ids'].append(layerCode)
                    rootNode['children'].append({
                        'name': item['name'],
                        'ids': [layerCode]
                    })
                else:
                    # Create root + register
                    rootNode = {
                        'name': item['parent'],
                        'ids': [ layerCode ],
                        'children': [
                            {
                                'name': item['name'],
                                'ids': [ layerCode ]
                            }
                        ]
                    }
                    parentNodes[item['parent']] = rootNode
                    pipelineMeta['pipeline'].append(rootNode)
            else:
                # Add item info as a new pipeline node
                pipelineMeta['pipeline'].append({
                    'name': item['name'],
                    'ids': [layerCode]
                })

        # Register metadata to be written in index.json
        self.dataHandler.addSection('Geometry', geometryMeta)
        self.dataHandler.addSection('CompositePipeline', pipelineMeta)

    def writeData(self, time=0):
        if not self.dataHandler.can_write:
            return

        currentScene = [];
        for data in self.config['scene']:
            currentData = {
                'name': data['name'],
                'fields': {},
                'cells': {}
            }
            currentScene.append(currentData)
            if self.surfaceExtract:
                self.merge.Input = data['source']
            else:
                self.merge = simple.MergeBlocks(Input=data['source'], MergePoints=0)
                self.surfaceExtract = simple.ExtractSurface(Input=self.merge)


            # Extract surface
            self.surfaceExtract.UpdatePipeline(time)
            ds = self.surfaceExtract.SMProxy.GetClientSideObject().GetOutputDataObject(0)
            originalDS = data['source'].SMProxy.GetClientSideObject().GetOutputDataObject(0)

            originalPoints = ds.GetPoints()

            # Points
            points = vtkFloatArray()
            nbPoints = originalPoints.GetNumberOfPoints()
            points.SetNumberOfComponents(3)
            points.SetNumberOfTuples(nbPoints)
            for idx in range(nbPoints):
                coord = originalPoints.GetPoint(idx)
                points.SetTuple3(idx, coord[0], coord[1], coord[2])

            pBuffer = buffer(points)
            pMd5 = hashlib.md5(pBuffer).hexdigest()
            pPath = os.path.join(self.dataHandler.getBasePath(), 'data',"%s.Float32Array" % pMd5)
            currentData['points'] = 'data/%s.Float32Array' % pMd5
            with open(pPath, 'wb') as f:
                f.write(pBuffer)

            # Handle cells
            writeCellArray(self.dataHandler, currentData['cells'], 'verts', ds.GetVerts().GetData())
            writeCellArray(self.dataHandler, currentData['cells'], 'lines', ds.GetLines().GetData())
            writeCellArray(self.dataHandler, currentData['cells'], 'polys', ds.GetPolys().GetData())
            writeCellArray(self.dataHandler, currentData['cells'], 'strips', ds.GetStrips().GetData())

            # Fields
            for fieldName, fieldInfo in iteritems(data['colors']):
                array = None
                if 'constant' in fieldInfo:
                    currentData['fields'][fieldName] = fieldInfo
                    continue
                elif 'POINT_DATA' in fieldInfo['location']:
                    array = ds.GetPointData().GetArray(fieldName)
                elif 'CELL_DATA' in fieldInfo['location']:
                    array = ds.GetCellData().GetArray(fieldName)
                jsType = jsMapping[arrayTypesMapping[array.GetDataType()]]
                arrayRange = array.GetRange(-1)
                tupleSize = array.GetNumberOfComponents()
                arraySize = array.GetNumberOfTuples()
                if tupleSize == 1:
                    outputField = array
                else:
                    # compute magnitude
                    outputField = array.NewInstance()
                    outputField.SetNumberOfTuples(arraySize)
                    tupleIdxs = range(tupleSize)
                    for i in range(arraySize):
                        magnitude = 0
                        for j in tupleIdxs:
                            magnitude += math.pow(array.GetValue(i * tupleSize + j), 2)

                        outputField.SetValue(i, math.sqrt(magnitude))

                fBuffer = buffer(outputField)
                fMd5 = hashlib.md5(fBuffer).hexdigest()
                fPath = os.path.join(self.dataHandler.getBasePath(), 'data',"%s_%s.%s" % (fieldName, fMd5, jsType))
                with open(fPath, 'wb') as f:
                    f.write(fBuffer)

                currentRange = self.ranges[fieldName]
                if currentRange[1] < currentRange[0]:
                    currentRange[0] = arrayRange[0];
                    currentRange[1] = arrayRange[1];
                else:
                    currentRange[0] = arrayRange[0] if arrayRange[0] < currentRange[0] else currentRange[0];
                    currentRange[1] = arrayRange[1] if arrayRange[1] > currentRange[1] else currentRange[1];

                currentData['fields'][fieldName] = {
                    'array' : 'data/%s_%s.%s' % (fieldName, fMd5, jsType),
                    'location': fieldInfo['location'],
                    'range': outputField.GetRange()
                }

        # Write scene
        with open(self.dataHandler.getDataAbsoluteFilePath('scene'), 'w') as f:
            f.write(json.dumps(currentScene, indent=4))


    def stop(self, compress=True, clean=True):
        if not self.dataHandler.can_write:
            return

        DataSetBuilder.stop(self)

        if compress:
            for dirName in ['fields', 'index', 'points']:
                for root, dirs, files in os.walk(os.path.join(self.dataHandler.getBasePath(), dirName)):
                    print ('Compress', root)
                    for name in files:
                        if 'Array' in name and '.gz' not in name:
                            with open(os.path.join(root, name), 'rb') as f_in:
                                with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                    shutil.copyfileobj(f_in, f_out)
                            os.remove(os.path.join(root, name))

# -----------------------------------------------------------------------------
# GeometryDataSetBuilder Dataset Builder
# -----------------------------------------------------------------------------

class GeometryDataSetBuilder(DataSetBuilder):
    def __init__(self, location, sceneConfig, metadata={}, sections={}):
        DataSetBuilder.__init__(self, location, None, metadata, sections)

        # Update data type
        self.dataHandler.addTypes('geometry');

        # Create a representation for all scene sources
        self.config = sceneConfig

        # Processing pipeline
        self.surfaceExtract = None

        # Add directory path
        self.dataHandler.registerData(priority=0, name='scene', rootFile=True, fileName='scene.json', type='json')

        # Create directory containers
        pointsPath = os.path.join(location, 'points')
        polyPath = os.path.join(location, 'index')
        colorPath = os.path.join(location, 'fields')
        for p in [pointsPath, polyPath, colorPath]:
            if not os.path.exists(p):
                os.makedirs(p)

        # Create metadata structure
        colorToCodeMap = {}
        parentNodes = {}
        pipelineMeta = {
            'layers': [],
            'pipeline': [],
            'layer_fields': {},
            'fields': {}
        }
        geometryMeta = {
            'ranges': {},
            'layer_map': {},
            'object_size': {}
        }
        self.objSize = geometryMeta['object_size']
        for item in sceneConfig['scene']:
            # Handle layer
            layerCode = encode_codes[len(pipelineMeta['layers'])]
            pipelineMeta['layers'].append(layerCode)
            geometryMeta['layer_map'][layerCode] = item['name']
            geometryMeta['object_size'][item['name']] = { 'points': 0, 'index': 0 }

            # Handle colors
            pipelineMeta['layer_fields'][layerCode] = []
            for fieldName in item['colors']:
                colorCode = None
                if fieldName in colorToCodeMap:
                    colorCode = colorToCodeMap[fieldName]
                else:
                    colorCode = encode_codes[len(colorToCodeMap)]
                    colorToCodeMap[fieldName] = colorCode
                    geometryMeta['ranges'][fieldName] = [0, 1] # FIXME we don't know the range

                pipelineMeta['layer_fields'][layerCode].append(colorCode)
                pipelineMeta['fields'][colorCode] = fieldName

            # Handle pipeline
            if 'parent' in item:
                # Need to handle hierarchy
                if item['parent'] in parentNodes:
                    # Fill children
                    rootNode = parentNodes[item['parent']]
                    rootNode['ids'].append(layerCode)
                    rootNode['children'].append({
                        'name': item['name'],
                        'ids': [layerCode]
                    })
                else:
                    # Create root + register
                    rootNode = {
                        'name': item['parent'],
                        'ids': [ layerCode ],
                        'children': [
                            {
                                'name': item['name'],
                                'ids': [ layerCode ]
                            }
                        ]
                    }
                    parentNodes[item['parent']] = rootNode
                    pipelineMeta['pipeline'].append(rootNode)
            else:
                # Add item info as a new pipeline node
                pipelineMeta['pipeline'].append({
                    'name': item['name'],
                    'ids': [layerCode]
                })

        # Register metadata to be written in index.json
        self.dataHandler.addSection('Geometry', geometryMeta)
        self.dataHandler.addSection('CompositePipeline', pipelineMeta)

    def writeData(self, time=0):
        if not self.dataHandler.can_write:
            return

        currentScene = [];
        for data in self.config['scene']:
            currentData = {
                'name': data['name'],
                'fields': {}
            }
            currentScene.append(currentData)
            if self.surfaceExtract:
                self.merge.Input = data['source']
            else:
                self.merge = simple.MergeBlocks(Input=data['source'], MergePoints=0)
                self.surfaceExtract = simple.ExtractSurface(Input=self.merge)


            # Extract surface
            self.surfaceExtract.UpdatePipeline(time)
            ds = self.surfaceExtract.SMProxy.GetClientSideObject().GetOutputDataObject(0)
            originalDS = data['source'].SMProxy.GetClientSideObject().GetOutputDataObject(0)

            originalPoints = ds.GetPoints()

            # Points
            points = vtkFloatArray()
            nbPoints = originalPoints.GetNumberOfPoints()
            points.SetNumberOfComponents(3)
            points.SetNumberOfTuples(nbPoints)
            for idx in range(nbPoints):
                coord = originalPoints.GetPoint(idx)
                points.SetTuple3(idx, coord[0], coord[1], coord[2])

            pBuffer = buffer(points)
            pMd5 = hashlib.md5(pBuffer).hexdigest()
            pPath = os.path.join(self.dataHandler.getBasePath(), 'points',"%s.Float32Array" % pMd5)
            currentData['points'] = 'points/%s.Float32Array' % pMd5
            with open(pPath, 'wb') as f:
                f.write(pBuffer)

            # Polys
            poly = ds.GetPolys()
            nbCells = poly.GetNumberOfCells()
            cellLocation = 0
            idList = vtkIdList()
            topo = vtkTypeUInt32Array()
            topo.Allocate(poly.GetData().GetNumberOfTuples())

            for cellIdx in range(nbCells):
                poly.GetCell(cellLocation, idList)
                cellSize = idList.GetNumberOfIds()
                cellLocation += cellSize + 1
                if cellSize == 3:
                    topo.InsertNextValue(idList.GetId(0))
                    topo.InsertNextValue(idList.GetId(1))
                    topo.InsertNextValue(idList.GetId(2))
                elif cellSize == 4:
                    topo.InsertNextValue(idList.GetId(0))
                    topo.InsertNextValue(idList.GetId(1))
                    topo.InsertNextValue(idList.GetId(3))
                    topo.InsertNextValue(idList.GetId(1))
                    topo.InsertNextValue(idList.GetId(2))
                    topo.InsertNextValue(idList.GetId(3))
                else:
                    print ("Cell size of", cellSize, "not supported")

            iBuffer = buffer(topo)
            iMd5 = hashlib.md5(iBuffer).hexdigest()
            iPath = os.path.join(self.dataHandler.getBasePath(), 'index',"%s.Uint32Array" % iMd5)
            currentData['index'] = 'index/%s.Uint32Array' % iMd5
            with open(iPath, 'wb') as f:
                f.write(iBuffer)

            # Grow object side
            self.objSize[data['name']]['points'] = max(self.objSize[data['name']]['points'], nbPoints)
            self.objSize[data['name']]['index'] = max(self.objSize[data['name']]['index'], topo.GetNumberOfTuples())

            # Colors / FIXME
            for fieldName, fieldInfo in iteritems(data['colors']):
                array = ds.GetPointData().GetArray(fieldName)
                tupleSize = array.GetNumberOfComponents()
                arraySize = array.GetNumberOfTuples()
                outputField = vtkFloatArray()
                outputField.SetNumberOfTuples(arraySize)
                if tupleSize == 1:
                    for i in range(arraySize):
                        outputField.SetValue(i, array.GetValue(i))
                else:
                    # compute magnitude
                    tupleIdxs = range(tupleSize)
                    for i in range(arraySize):
                        magnitude = 0
                        for j in tupleIdxs:
                            magnitude += math.pow(array.GetValue(i * tupleSize + j), 2)

                        outputField.SetValue(i, math.sqrt(magnitude))

                fBuffer = buffer(outputField)
                fMd5 = hashlib.md5(fBuffer).hexdigest()
                fPath = os.path.join(self.dataHandler.getBasePath(), 'fields',"%s_%s.Float32Array" % (fieldName, fMd5))
                with open(fPath, 'wb') as f:
                    f.write(fBuffer)

                currentData['fields'][fieldName] = 'fields/%s_%s.Float32Array' % (fieldName, fMd5)

        # Write scene
        with open(self.dataHandler.getDataAbsoluteFilePath('scene'), 'w') as f:
            f.write(json.dumps(currentScene, indent=2))


    def stop(self, compress=True, clean=True):
        if not self.dataHandler.can_write:
            return

        DataSetBuilder.stop(self)

        if compress:
            for dirName in ['fields', 'index', 'points']:
                for root, dirs, files in os.walk(os.path.join(self.dataHandler.getBasePath(), dirName)):
                    print ('Compress', root)
                    for name in files:
                        if 'Array' in name and '.gz' not in name:
                            with open(os.path.join(root, name), 'rb') as f_in:
                                with gzip.open(os.path.join(root, name + '.gz'), 'wb') as f_out:
                                    shutil.copyfileobj(f_in, f_out)
                            os.remove(os.path.join(root, name))
